"""
"""
from __future__ import annotations

import contextlib
import json
import os
from contextvars import ContextVar
from io import BytesIO
from pathlib import Path
from typing import Any
from typing import Callable
from typing import cast
from unittest.mock import patch

import torch
from packaging import version
if version.parse(torch.__version__) < version.parse('2.8'): # pragma: no cover
    raise RuntimeError("ZeroGPU AoTI reuqires PyTorch 2.8+")

from torch._functorch._aot_autograd.subclass_parametrization import unwrap_tensor_subclass_parameters
from torch._inductor.package.package import package_aoti
from torch.export.pt2_archive._package import AOTICompiledModel
from torch.export.pt2_archive._package_weights import Weights

from ..utils import register_cleanup


INDUCTOR_CONFIGS_OVERRIDES = {
    'aot_inductor.package_constants_in_so': False,
    'aot_inductor.package_constants_on_disk': True,
    'aot_inductor.package': True,
    'always_keep_tensor_constants': True,
}


ARCHIVE_SO_PATTERN = '/tmp/*/archive/data/aotinductor/model/*.wrapper.so'

PACKAGE_FILENAME = 'package.pt2'


@contextlib.contextmanager
def _register_aoti_cleanup():
    """
    PyTorch already cleans-up extracted archives in /tmp
    But the GPU worker never terminates gracefully in ZeroGPU so cleanup must be done manually
    """
    pid = os.getpid()
    map_files = Path(f'/proc/{pid}/map_files')
    maps_before = {f.name for f in map_files.iterdir()}
    yield
    for map_file in map_files.iterdir():
        if map_file.name not in maps_before:
            if (mapped := map_file.readlink()).match(ARCHIVE_SO_PATTERN):
                package_path = Path(*mapped.parts[:3])
                return register_cleanup(pid, package_path)


class ZeroGPUWeights:
    def __init__(self, constants_map: dict[str, torch.Tensor], to_cuda: bool = False):
        if to_cuda:
            self.constants_map = {name: tensor.to('cuda') for name, tensor in constants_map.items()}
        else:
            self.constants_map = constants_map
    def __reduce__(self):
        constants_map: dict[str, torch.Tensor] = {}
        for name, tensor in self.constants_map.items():
            tensor_ = torch.empty_like(tensor, device='cpu').pin_memory()
            constants_map[name] = tensor_.copy_(tensor).detach().share_memory_()
        return ZeroGPUWeights, (constants_map, True)


class ZeroGPUCompiledModel:
    def __init__(self, archive_file: torch.types.FileLike, weights: ZeroGPUWeights):
        self.archive_file = archive_file
        self.weights = weights
        self.compiled_model: ContextVar[AOTICompiledModel | None] = ContextVar('compiled_model', default=None)
    def __call__(self, *args, **kwargs):
        if (compiled_model := self.compiled_model.get()) is None:
            with _register_aoti_cleanup():
                compiled_model = torch._inductor.aoti_load_package(self.archive_file)
            compiled_model = cast(AOTICompiledModel, compiled_model)
            constant_map = {name: self.weights.constants_map[name] for name in compiled_model.get_constant_fqns()}
            compiled_model.load_constants(constant_map, check_full_update=True, user_managed=True)
            self.compiled_model.set(compiled_model)
        return compiled_model(*args, **kwargs)
    def __reduce__(self):
        return ZeroGPUCompiledModel, (self.archive_file, self.weights)


def aoti_compile(
    exported_program: torch.export.ExportedProgram,
    inductor_configs: dict[str, Any] | None = None,
):
    inductor_configs = {**(inductor_configs or {}), **INDUCTOR_CONFIGS_OVERRIDES}
    gm = cast(torch.fx.GraphModule, exported_program.module())
    assert exported_program.example_inputs is not None
    args, kwargs = exported_program.example_inputs
    artifacts = torch._inductor.aot_compile(gm, args, kwargs, options=inductor_configs) # pyright: ignore [reportArgumentType]
    artifacts = cast(list[str | Weights], artifacts)
    archive_file = BytesIO()
    files = (file for file in artifacts if isinstance(file, str))
    package_aoti(archive_file, list(files))
    weights, = (artifact for artifact in artifacts if isinstance(artifact, Weights))
    weights = cast(Weights, weights)
    zerogpu_weights = ZeroGPUWeights({name: weights.get_weight(name)[0] for name in weights})
    return ZeroGPUCompiledModel(archive_file, zerogpu_weights)


def aoti_apply(
    compiled: ZeroGPUCompiledModel,
    module: torch.nn.Module,
    call_method: str = 'forward',
):
    setattr(module, call_method, compiled)
    drain_module_parameters(module)


def drain_module_parameters(module: torch.nn.Module):
    state_dict_meta = {name: {'device': tensor.device, 'dtype': tensor.dtype} for name, tensor in module.state_dict().items()}
    state_dict = {name: torch.nn.Parameter(torch.empty_like(tensor, device='cpu')) for name, tensor in module.state_dict().items()}
    module.load_state_dict(state_dict, assign=True)
    for name, param in state_dict.items():
        meta = state_dict_meta[name]
        param.data = torch.Tensor([]).to(**meta)


@contextlib.contextmanager
def aoti_capture(
    module: torch.nn.Module | Callable[..., Any],
    call_method: str = 'forward',
):
    class CapturedCallException(Exception):
        def __init__(self, *args, **kwargs):
            super().__init__()
            self.args = args
            self.kwargs = kwargs

    class CapturedCall:
        def __init__(self):
            self.args: tuple[Any, ...] = ()
            self.kwargs: dict[str, Any] = {}

    captured_call = CapturedCall()

    def capture_call(*args, **kwargs):
        raise CapturedCallException(*args, **kwargs)

    with patch.object(module, call_method, new=capture_call):
        try:
            yield captured_call
        except CapturedCallException as e:
            captured_call.args = e.args
            captured_call.kwargs = e.kwargs


class LazyAOTIModel:
    def __init__(self, archive_file: torch.types.FileLike):
        self.archive_file = archive_file
        self.compiled_model: ContextVar[AOTICompiledModel | None] = ContextVar('compiled_model', default=None)
        self.loaded_weights: ContextVar[dict[str, torch.Tensor] | None] = ContextVar('loaded_weights', default=None)
    def __call__(self, weights: dict[str, torch.Tensor], check_full_update: bool, *args, **kwargs):
        if (compiled_model := self.compiled_model.get()) is None:
            with _register_aoti_cleanup():
                compiled_model = torch._inductor.aoti_load_package(self.archive_file)
            compiled_model = cast(AOTICompiledModel, compiled_model)
            self.compiled_model.set(compiled_model)
        if (loaded_weights := self.loaded_weights.get()) is None or loaded_weights is not weights:
            compiled_model.load_constants(weights, check_full_update=check_full_update, user_managed=True)
            self.loaded_weights.set(weights)
        return compiled_model(*args, **kwargs)
    def with_weights(self, weights: dict[str, torch.Tensor]):
        return LazyAOTIModelWithWeights(self, weights)


class LazyAOTIModelWithWeights:
    def __init__(self, model: LazyAOTIModel, weights: dict[str, torch.Tensor]):
        self.model = model
        self.weights = weights
        self.first_call = True
    def __call__(self, *args, **kwargs):
        check_full_update = self.first_call
        self.first_call = False
        return self.model(self.weights, check_full_update, *args, **kwargs)


def _shallow_clone_module(module: torch.nn.Module) -> torch.nn.Module:
    clone = object.__new__(module.__class__)
    clone.__dict__ = module.__dict__.copy()
    clone._parameters = module._parameters.copy()
    clone._buffers = module._buffers.copy()
    clone._modules = {k: _shallow_clone_module(v) for k, v in module._modules.items() if v is not None}
    return clone


def aoti_blocks_load(module: torch.nn.Module, repo_id: str, variant: str | None = None):
    """
    Loads AOTI-compiled blocks for a given module from the Hugging Face Hub.
    This function expects the module to expose a `_repeated_blocks` attribute.
    This attribute is present on most models from the diffusers library.

    Args:
        module (torch.nn.Module): The module containing repeated blocks to be replaced.
        repo_id (str): The Hugging Face Hub repository ID where the compiled blocks are stored.
        variant (str | None, optional): Variant suffix to append to block names. Defaults to None.

    Returns:
        None: The function mutates the given module in place.

    Example:
        >>> import spaces
        >>> import torch
        >>> from diffusers import FluxPipeline
        >>> pipeline = FluxPipeline.from_pretrained('black-forest-labs/FLUX.1-dev')
        >>> spaces.aoti_blocks_load(pipeline.transformer, 'zerogpu-aoti/FLUX.1')
    """

    from huggingface_hub import hf_hub_download
    from huggingface_hub.errors import EntryNotFoundError

    if (repeated_blocks := getattr(module, '_repeated_blocks', None)) is None:
        raise RuntimeError("aoti_blocks_load only works with modules that expose _repeated_blocks")

    #region: Quick and dirty config support
    try:
        config_path = hf_hub_download(repo_id, f"{_variant('config', variant)}.json")
    except EntryNotFoundError:
        config_path = None
    config = json.loads(Path(config_path).read_text() if config_path is not None else '{}')
    if (kernels_config := config.get('kernels', None)) is not None:
        for kernels_kwargs in kernels_config:
            from kernels import get_kernel
            get_kernel(**kernels_kwargs) # Load custom ops
    #endregion

    repeated_blocks = cast(list[str], repeated_blocks)
    aoti_models = {name: LazyAOTIModel(hf_hub_download(
        repo_id=repo_id,
        filename=PACKAGE_FILENAME,
        subfolder=_variant(name, variant),
    )) for name in repeated_blocks}

    for block_name, aoti_model in aoti_models.items():
        for block in module.modules():
            if block.__class__.__name__ == block_name:
                block_ = _shallow_clone_module(block) # Prevent original block mutation
                unwrap_tensor_subclass_parameters(block_) # https://github.com/pytorch/pytorch/issues/159918
                block.forward = aoti_model.with_weights(block_.state_dict())


def _variant(name: str, variant: str | None):
    return f'{name}.{variant}' if variant else name



"""
"""
from __future__ import annotations

import inspect
from functools import wraps
from packaging import version
from typing import Any
from typing import Callable
from typing import Literal
from typing import NamedTuple
from typing import TYPE_CHECKING
import warnings

import gradio as gr
from gradio.context import Context
from gradio.context import LocalContext
from gradio.helpers import Progress
from gradio.helpers import TrackedIterable
from gradio.queueing import Queue
from typing_extensions import ParamSpec
from typing_extensions import TypeAlias

from ..utils import SimpleQueue
from .types import GeneratorResQueueResult
from .types import GradioQueueEvent
from .types import RegularResQueueResult


QUEUE_RPC_METHODS = [
    "set_progress",
    "log_message",
]


try:
    Success = gr.Success # pyright: ignore[reportAttributeAccessIssue] (Gradio<5.10)
except AttributeError: # pragma: no cover
    Success = gr.Info

Level: TypeAlias = "Literal['success', 'info', 'warning']"

def modal(level: Level):
    if level == 'info':
        return gr.Info
    if level == 'success':
        return Success
    if level == 'warning':
        return gr.Warning


class GradioPartialContext(NamedTuple):
    event_id: str | None
    in_event_listener: bool
    progress: Progress | None

    @staticmethod
    def get():
        TrackedIterable.__reduce__ = tracked_iterable__reduce__
        return GradioPartialContext(
            event_id=LocalContext.event_id.get(None),
            in_event_listener=LocalContext.in_event_listener.get(False),
            progress=LocalContext.progress.get(None),
        )

    @staticmethod
    def apply(context: 'GradioPartialContext'):
        LocalContext.event_id.set(context.event_id)
        LocalContext.in_event_listener.set(context.in_event_listener)
        LocalContext.progress.set(context.progress)


def get_queue_instance():
    blocks = LocalContext.blocks.get(None)
    if blocks is None: # pragma: no cover
        return None
    return blocks._queue


def get_event():
    queue = get_queue_instance()
    event_id = LocalContext.event_id.get(None)
    if queue is None:
        return None
    if event_id is None: # pragma: no cover
        return None
    for job in queue.active_jobs:
        if job is None: # pragma: no cover
            continue
        for event in job:
            if event._id == event_id:
                return event


def get_server_port() -> int | None:
    from_request_context = True
    if (blocks := LocalContext.blocks.get(None)) is None: # Request
        from_request_context = False
        if (blocks := Context.root_block) is None: # Caching
            return None
    if (server := getattr(blocks, 'server', None)) is None: # pragma: no cover (Gradio 4)
        if from_request_context:
            warnings.warn("Gradio: No blocks.server inside a request") # pragma: no cover
        return -1
    if TYPE_CHECKING:
        assert (server := blocks.server)
    return server.config.port


def try_process_queue_event(method_name: str, *args, **kwargs):
    queue = get_queue_instance()
    if queue is None: # pragma: no cover
        warnings.warn("ZeroGPU: Cannot get Gradio app Queue instance")
        return
    method = getattr(queue, method_name, None)
    assert callable(method)
    method(*args, **kwargs)


def patch_gradio_queue(
    res_queue: SimpleQueue[RegularResQueueResult | None] | SimpleQueue[GeneratorResQueueResult | None],
):

    def rpc_method(method_name: str):
        def method(*args, **kwargs):
            if args and isinstance(args[0], Queue):
                args = args[1:] # drop `self`
            res_queue.put(GradioQueueEvent(method_name, args, kwargs))
        return method

    for method_name in QUEUE_RPC_METHODS:
        if (method := getattr(Queue, method_name, None)) is None: # pragma: no cover
            warnings.warn(f"ZeroGPU: Gradio Queue has no {method_name} attribute")
            continue
        if not callable(method): # pragma: no cover
            warnings.warn(f"ZeroGPU: Gradio Queue {method_name} is not callable")
            continue
        setattr(Queue, method_name, rpc_method(method_name))

    TrackedIterable.__reduce__ = tracked_iterable__reduce__


def tracked_iterable__reduce__(self):
    res: tuple = super(TrackedIterable, self).__reduce__() # type: ignore
    cls, base, state, *_ = res
    return cls, base,{**state, **{
        'iterable': None,
        '_tqdm': None,
    }}


def supports_auth():
    return version.parse(gr.__version__) >= version.Version('4.27.0')


Param = ParamSpec('Param')

_orig_launch = gr.Blocks.launch

def one_launch(task: Callable[Param, None], *task_args: Param.args, **task_kwargs: Param.kwargs):
    @wraps(gr.Blocks.launch)
    def launch(*args, **kwargs):
        task(*task_args, **task_kwargs)
        gr.Blocks.launch = _orig_launch
        return gr.Blocks.launch(*args, **kwargs)
    gr.Blocks.launch = launch


class HTMLError(gr.Error):
    def __str__(self): # pragma: no cover
        return str(self.message)


def error(title: str, message: str, html: bool = False):
    params = inspect.signature(gr.Error).parameters
    kwargs: dict[str, Any] = {}
    if 'title' in params:
        kwargs = {**kwargs, 'title': title}
    if 'print_exception' in params:
        kwargs = {**kwargs, 'print_exception': False}
    error_cls = HTMLError if html else gr.Error
    return error_cls(message, **kwargs)


def info(title: str, message: str, level: Level = 'info'):
    params = inspect.signature(gr.Info).parameters
    kwargs: dict[str, Any] = {}
    if 'title' in params:
        kwargs = {**kwargs, 'title': title}
    info_cls = modal(level)
    return info_cls(message, **kwargs)



"""
"""
from __future__ import annotations

import inspect
import sys
import warnings
from datetime import timedelta
from functools import partial
from typing import Callable
from typing import TypeVar
from typing import overload
from typing_extensions import ParamSpec
from typing_extensions import Unpack

from ..config import Config
from .types import DynamicDuration
from .types import EmptyKwargs


P = ParamSpec('P')
R = TypeVar('R')


decorated_cache: dict[Callable, Callable] = {}


@overload
def GPU(
    task: None = None, *,
    duration: DynamicDuration[P] = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    ...
@overload
def GPU(
    task: Callable[P, R], *,
    duration: DynamicDuration[P] = None,
) -> Callable[P, R]:
    ...
def GPU(
    task: Callable[P, R] | None = None, *,
    duration: DynamicDuration[P] = None,
    **kwargs: Unpack[EmptyKwargs],
) -> Callable[[Callable[P, R]], Callable[P, R]] | Callable[P, R]:
    """
    ZeroGPU decorator

    Basic usage:
        ```
        @spaces.GPU
        def fn(...):
            # CUDA is available here
            pass
        ```

    With custom duration:
        ```
        @spaces.GPU(duration=45) # Expressed in seconds
        def fn(...):
            # CUDA is available here
            pass
        ```

    Args:
        task (`Callable | None`): Python function that requires CUDA
        duration (`int | datetime.timedelta`): Estimated duration in seconds or `datetime.timedelta`

    Returns:
        `Callable`: GPU-ready function
    """
    if "enable_queue" in kwargs:
        warnings.warn("`enable_queue` parameter is now ignored and always set to `True`")
    if task is None:
        return partial(_GPU, duration=duration)
    return _GPU(task, duration)


def _GPU(
    task: Callable[P, R],
    duration: DynamicDuration[P],
) -> Callable[P, R]:

    if not Config.zero_gpu:
        return task

    from . import client
    from .wrappers import regular_function_wrapper
    from .wrappers import generator_function_wrapper

    if sys.version_info.minor < 9: # pragma: no cover
        raise RuntimeError("Actually using @spaces.GPU on a ZeroGPU Space requires Python 3.9+")

    if task in decorated_cache:
        # TODO: Assert same duration ?
        return decorated_cache[task] # type: ignore

    if inspect.iscoroutinefunction(task):
        raise NotImplementedError

    if inspect.isgeneratorfunction(task):
        decorated = generator_function_wrapper(task, duration)
    else:
        decorated = regular_function_wrapper(task, duration)

    setattr(decorated, 'zerogpu', None)

    decorated_cache.update({
        task:      decorated,
        decorated: decorated,
    })

    return decorated # type: ignore



"""
"""
from __future__ import annotations

import traceback
import threading
from pathlib import Path
from queue import Queue
from typing import Any
from typing import Callable
from typing import NamedTuple
from typing_extensions import assert_type
from typing_extensions import Never
from uuid import uuid4

from fastapi import APIRouter
from fastapi import FastAPI
from fastapi import HTTPException
from uvicorn import Config
from uvicorn import Server

from .._vendor.jurigged.codetools import CodeFile
from .._vendor.jurigged.codetools import CodeFileOperation
from .._vendor.jurigged.codetools import AddOperation
from .._vendor.jurigged.codetools import UpdateOperation
from .._vendor.jurigged.codetools import DeleteOperation
from .._vendor.jurigged.codetools import ExceptionOperation
from .._vendor.jurigged.codetools import Extent
from .._vendor.jurigged.codetools import LineDefinition
from .._vendor.jurigged.codetools import GroupDefinition
from .._vendor.jurigged.codetools import ModuleCode
from .._vendor.jurigged.codetools import ClassDefinition
from .._vendor.jurigged.codetools import FunctionDefinition
from .._vendor.jurigged.register import EventSource
from .._vendor.jurigged.register import Registry
from .._vendor.jurigged.utils import glob_filter
from .._vendor.sse_starlette import EventSourceResponse
from ..utils import create_thread

from .types import ApiCreateReloadRequest
from .types import ApiCreateReloadResponse
from .types import ApiCreateReloadResponseError
from .types import ApiCreateReloadResponseSuccess
from .types import ApiGetReloadRequest
from .types import ApiGetReloadEventSourceData
from .types import ApiGetStatusRequest
from .types import ApiGetStatusResponse
from .types import ApiFetchContentsRequest
from .types import ApiFetchContentsResponse
from .types import ApiFetchContentsResponseError
from .types import ApiFetchContentsResponseSuccess
from .types import ReloadRegion
from .types import ReloadOperationError
from .types import ReloadOperationException
from .types import ReloadOperationObject
from .types import ReloadOperationRun
from .types import ReloadOperationUI


class ReloadRun(NamedTuple):
    activity: EventSource
    thread: threading.Thread


class ReloadServer:
    def __init__(
        self,
        prerun: Callable[[], Any],
        postrun: Callable[[], bool],
        stop_event: threading.Event,
    ):
        self.prerun = prerun
        self.postrun = postrun
        self.stop_event = stop_event
        self.create_lock = threading.Lock()
        self.reload_runs: dict[str, ReloadRun] = {}
        self.active_reload_id: str | None = None
        self.router = APIRouter()
        self.router.add_api_route('/healthz', lambda: {}, methods=['GET'])
        self.router.add_api_route('/get-status', self.get_status, methods=['POST'])
        self.router.add_api_route('/fetch-contents', self.fetch_contents, methods=['POST'])
        self.router.add_api_route('/create-reload', self.create_reload, methods=['POST'])
        self.router.add_api_route('/get-reload', self.get_reload, methods=['POST'])
        self.registry = Registry()
        self.registry.auto_register(filter=glob_filter('./*.py'))


    def get_status(self, req: ApiGetStatusRequest) -> ApiGetStatusResponse:
        """
        POST /get-status
        """
        raise NotImplementedError # pragma: no cover


    def fetch_contents(self, req: ApiFetchContentsRequest) -> ApiFetchContentsResponse:
        """
        POST /fetch-contents
        """

        filepath = (Path.cwd() / req.filepath).resolve()
        if not filepath.is_relative_to(Path.cwd()) or not (file := Path(req.filepath)).is_file():
            res = ApiFetchContentsResponseError(status='fileNotFound')
            return  ApiFetchContentsResponse(res=res)

        res = ApiFetchContentsResponseSuccess(status='ok', contents=file.read_text())
        return ApiFetchContentsResponse(res=res)


    def _create_reload(self, req: ApiCreateReloadRequest):

        if self.active_reload_id is not None:
            return  ApiCreateReloadResponseError(status='alreadyReloading')

        filepath = (Path.cwd() / req.filepath).resolve()
        if not filepath.is_relative_to(Path.cwd()) or (code_file := self.registry.get(str(filepath))) is None:
            return  ApiCreateReloadResponseError(status='fileNotFound')

        Path(req.filepath).write_text(req.contents)
        activity = EventSource(save_history=True) # TODO: Generic EventSource[ApiGetReloadEventSourceMessage]
        code_file.activity = activity
        thread = create_thread(self.run_reload, code_file)
        thread.start()

        reload_id = str(uuid4()) # TODO: f"{commit_sha}-{counter.next()}"
        self.reload_runs[reload_id] = ReloadRun(activity, thread)
        self.active_reload_id = reload_id

        return ApiCreateReloadResponseSuccess(status='created', reloadId=reload_id)


    def create_reload(self, req: ApiCreateReloadRequest) -> ApiCreateReloadResponse:
        """
        POST /create-reload
        """

        with self.create_lock:
            res = self._create_reload(req)

        return ApiCreateReloadResponse(res=res)


    def get_reload(self, req: ApiGetReloadRequest):
        """
        POST /get-reload
        """

        reload_id = req.reloadId
        if (run := self.reload_runs.get(reload_id, None)) is None:
            raise HTTPException(404, f"{reload_id=} not found")

        queue = Queue()
        run.activity.register(queue.put, apply_history=True)

        def activity_stream(queue: Queue):
            while (item := queue.get()) is not None:
                if isinstance(item, ApiGetReloadEventSourceData):
                    yield item
                elif isinstance(item, CodeFileOperation):
                    if (op := serialize_code_operation(item)) is not None:
                        yield ApiGetReloadEventSourceData(data=op)

        def event_source_stream():
            for data in activity_stream(queue):
                yield data.model_dump_json()

        return EventSourceResponse(event_source_stream())


    def run_reload(self, code_file: CodeFile):
        updated = False
        try:
            self.prerun()
            code_file.refresh()
            updated = self.postrun()
        except Exception as exc:
            tb = format_traceback(exc)
            op = ReloadOperationError(kind='error', traceback=tb)
            data = ApiGetReloadEventSourceData(data=op)
            code_file.activity.emit(data)
        op = ReloadOperationUI(kind='ui', updated=updated)
        data = ApiGetReloadEventSourceData(data=op)
        code_file.activity.emit(data)
        code_file.activity.emit(None)
        self.active_reload_id = None


    def run(self, port: int):
        app = FastAPI()
        app.include_router(self.router)
        server = Server(Config(app, port=port, log_level='warning'))
        server_thread = create_thread(server.run)
        server_thread.start()
        self.stop_event.wait()
        server.should_exit = True


def serialize_code_operation(cf_operation: CodeFileOperation):
    assert isinstance(defn := cf_operation.defn, (LineDefinition, GroupDefinition))
    region = serialize_extent(defn.stashed)
    if isinstance(cf_operation, ExceptionOperation):
        if (exc := cf_operation.exc) is None: # pragma: no cover
            exc = Exception("Unable to retrieve reload exception")
        tb = format_traceback(exc)
        return ReloadOperationException(kind='exception', region=region, traceback=tb)
    if isinstance(defn, GroupDefinition):
        otype = get_object_type(defn)
        oname = defn.dotpath()
        if isinstance(cf_operation, AddOperation):
            kind = 'add'
        elif isinstance(cf_operation, UpdateOperation):
            if not isinstance(defn, FunctionDefinition):
                return None
            kind = 'update'
        else:
            kind = 'delete'
        return ReloadOperationObject(kind=kind, region=region, objectType=otype, objectName=oname)
    if isinstance(defn, LineDefinition):
        if isinstance(cf_operation, DeleteOperation):
            return None
        return ReloadOperationRun(kind='run', region=region, codeLines=defn.text)
    assert_type(defn, Never) # pragma: no cover


def get_object_type(defn: GroupDefinition):
    if isinstance(defn, ClassDefinition):
        return 'class'
    if isinstance(defn, ModuleCode):
        return 'module'
    if isinstance(defn, FunctionDefinition):
        return 'function'
    return 'unknown' # pragma: no cover


def serialize_extent(extent: Extent):
    return ReloadRegion(
        startLine=extent.lineno,
        startCol=extent.col_offset,
        endLine=extent.end_lineno,
        endCol=extent.end_col_offset,
    )


def format_traceback(exc: Exception):
    traces = traceback.format_exception(type(exc), exc, exc.__traceback__)
    return "".join(traces)
